import os
import subprocess
from pathlib import Path
from typing import Sequence

import click
import shellingham
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.css.query import QueryError
from textual.widgets import Footer

from clippt.slides import Slide, load
from clippt.theming import css_tweaks, my_theme


class PresentationApp(App):
    """A Textual app for the presentation."""

    enable_footer: bool = True

    BINDINGS = [
        ("pageup", "prev_slide", "Previous"),
        ("pagedown", "next_slide", "Next"),
        (".", "run", "Run"),
        ("q", "quit", "Quit"),
        ("e", "edit", "Edit"),
        ("r", "reload", "Reload"),
        ("home", "home", "First slide"),
        ("end", "end", "Last slide"),
        ("ctrl+o", "shell", "Shell"),
    ]

    CSS = css_tweaks

    slide_index: int = 0

    slides: list[Slide]

    document_title: str

    shell_cwd: Path | None = None

    def __init__(self, *, slides: Sequence[str | Path | Slide], title: str, shell_cwd: Path | None = None, **kwargs):
        self.slides = self._ensure_load_slides(list(slides))
        self.document_title = title
        self.shell_cwd = shell_cwd
        super().__init__(**kwargs)

    def _ensure_load_slides(self, slides: list[Slide | str | Path]) -> list[Slide]:
        return [
            slide_or_path if isinstance(slide_or_path, Slide) else load(slide_or_path)
            for slide_or_path in slides
        ]

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        # yield Header(show_clock=True)
        yield Container(
            self.current_slide.render(app=self),
            id="content",  # , can_focus=False
        )
        if self.enable_footer:
            yield Footer(show_command_palette=False)

    def on_mount(self) -> None:
        """Hook called when the app is mounted."""
        self.register_theme(my_theme)
        self.theme = "my"

    def on_resize(self) -> None:
        """Hook called when the app is resized."""
        self.update_slide()

    def action_reload(self) -> None:
        self.current_slide.reload()
        self.update_slide()

    def action_next_slide(self) -> None:
        self.switch_to_slide(min(self.slide_index + 1, len(self.slides) - 1))

    def action_prev_slide(self) -> None:
        self.switch_to_slide(max(self.slide_index - 1, 0))

    def action_home(self) -> None:
        self.switch_to_slide(0)

    def action_end(self) -> None:
        self.switch_to_slide(len(self.slides) - 1)

    def switch_to_slide(self, index: int) -> None:
        curent_index = self.slide_index
        if index != curent_index:
            self.slide_index = index
            self.update_slide()

    def action_edit(self) -> None:
        if self.current_slide.path:
            with self.suspend():
                click.edit(
                    filename=[str(self.current_slide.path)],
                    editor=os.environ.get("EDITOR"),
                )
            self.current_slide.reload()
        self.update_slide()

    def action_run(self) -> None:
        if self.current_slide.runnable:
            self.current_slide.run()
            self.update_slide()
            # No need to refresh() - update_slide() handles the refresh

    @property
    def current_slide(self) -> Slide:
        return self.slides[self.slide_index]

    def update_slide(self) -> None:
        try:
            container_widget = self.query_one("#content", Container)
            if not container_widget.is_attached:
                return
            container_widget.remove_children()
            # No need to refresh() - mounting will trigger automatic refresh
            self.log("Rendering slide", self.current_slide.model_dump())
            content_widget = self.slides[self.slide_index].render(app=self)
            container_widget.mount(content_widget)
            Path(".current_slide").write_text(str(self.slide_index))
        except QueryError:
            pass

    def action_shell(self):
        with self.suspend():
            _, shell = shellingham.detect_shell()
            subprocess.run(shell, shell=True, capture_output=False, cwd=self.shell_cwd)



from textual.theme import Theme

my_theme = Theme(
    name="my",
    primary="#0000c0",
    secondary="#4040ff",
    accent="#00ff00",
    foreground="#444444",
    background="#ffffff",
    success="#A3BE8C",
    warning="#EBCB8B",
    error="#BF616A",
    surface="#ffffff",
    panel="#ffffff",
    dark=False,
    variables={
        "block-cursor-text-style": "none",
        "footer-key-foreground": "#88C0D0",
        "input-selection-background": "#81a1c1 35%",
    },
)

css_tweaks = """
    Screen {
        align: center middle;
    }
    Footer {
        background: #ffffd0;
        color: white;
        dock: bottom;
        height: 1;
    }
    Markdown.title {
        MarkdownFence {
            background: white;
            margin: 0;
        }
    }
    MarkdownFence {
        background: #fff9e4; /* Light yellow */
        margin: 1;
        max-height: 500;
    }
    Static.output {
        background: #ccccff;
        margin: 0 3;
        max-height: 500;
        padding: 1 2;
    }
    Static.error {
        background: #ffcccc;
        margin: 0 3;
        max-height: 500;
        padding: 1 2;
    }
 """



"""Create a textual"""

from clippt.app import PresentationApp
from clippt.slides import load, md, py, sh, slide

__all__ = [
    "PresentationApp",
    "md",
    "sh",
    "py",
    "load",
    "slide",
]



import os
import sys

import rich


def wait_for_key():
    """Wait for a key press from the user."""
    rich.print("[bold]Press any key to continue...[/bold]")
    match sys.platform:
        case "win32":
            import msvcrt

            msvcrt.getch()
        case "linux" | "darwin":
            # Test this works on mac
            import termios
            import tty

            old_settings = termios.tcgetattr(sys.stdin)
            tty.setcbreak(sys.stdin.fileno())
            try:
                os.read(sys.stdin.fileno(), 3).decode()
            finally:
                termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)



from pathlib import Path

import click

from clippt.app import PresentationApp
from clippt.presentation import load_presentation, Presentation


@click.command()
@click.argument(
    "source",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        path_type=Path,
    ),
)
@click.option("--disable-footer", is_flag=True, help="Disable footer.")
@click.option(
    "--continue", "-c", "continue_", is_flag=True, help="Continue from last slide."
)
def clippt(*, source: Path, disable_footer: bool, continue_: bool):
    """Run a presentation in the command-line."""

    presentation = create_presentation(source)
    slides = list(presentation.create_slides())
    app = PresentationApp(slides=slides, title=presentation.title or "", shell_cwd=presentation.shell_cwd)
    app.enable_footer = not disable_footer
    if continue_ and Path(".current_slide").exists():
        app.slide_index = int(Path(".current_slide").read_text())
    app.slide_index = min(app.slide_index, len(slides) - 1)
    app.run()


def create_presentation(source: Path) -> Presentation:
    if source.is_dir():
        if (source / "presentation.toml").exists():
            source = source / "presentation.toml"
        else:
            raise FileNotFoundError(f"presentation.toml not found in {source}")

    match source.suffix.lower():
        case ".md" | ".markdown" | ".csv" | ".py" | ".pq" | ".parquet":
            return Presentation(slides=[str(source)])
        case ".toml":
            presentation = load_presentation(source)
            return presentation
        case _:
            raise ValueError(f"Unsupported file type: {source.suffix}")


if __name__ == "__main__":
    clippt()



from io import StringIO

from clippt.app import PresentationApp
from clippt.presentation import load_presentation

CONTENT = """
title = "My Presentation"
slides = [
    {source = "# This is slide 1"},
    {source = "import this", type = "python" },
]
"""

if __name__ == "__main__":
    presentation = load_presentation(StringIO(CONTENT))
    slides = list(presentation.create_slides())
    app = PresentationApp(slides=slides, title=presentation.title or "")
    app.run()



# /// script
# requires-python = ">=3.14"
# dependencies = [
#     "clippt>=0.1.4",
# ]
# ///
from io import StringIO

from clippt.app import PresentationApp
from clippt.presentation import load_presentation

CONTENT = '''
title = "Pražské Pyvo #177"

[[slides]]
source = """
<!-- date: 18/2/2026 -->
# Pražské Pyvo #177
Programme starts 19:00
"""

[[slides]]
source = """
# Start moved!
- Why? For many people start (and end) was too late.
- How much: half an hour sooner: 18:30 is door opened.
- Programme starts at 19:00 now (30 to 60 minutes sooner than usual)
"""

[[slides]]
source = """
# House rules
"""

[[slides]]
source = """
# Go outside to the garden only ⬆
- neighbour relations 🥚🔇
- if not leaving
"""

[[slides]]
source = """
# Don't forget to pay before leaving 💸
"""

[[slides]]
source = """
# Cash only 💰
If you don't have cash on you ask organizers.
"""

[[slides]]
source = """
# Be nice to each other
Everybody should feel welcome and safe. Contact organizers if you're not.
"""

[[slides]]
source = """
# Program

18:30 – Doors open
19:00 – 🗣️ Talk "Cyber Resilience Act and Open-Source"
19:45 – ⚡ Lightning talks: add yours to https://bit.ly/prpylight
after 20:00 – 👯 Chat, singing etc. until midnight
"""

[[slides]]
source = """
# Lightning talks
- any topic
- 5 minutes per talk maximum
- sign up bit.ly/prpylight
"""

[[slides]]
source = """
# Introductions
Please tell everybody **Try to speak **LOUDLY** in CZ, SK or EN.**
1. Your name
2. Did you ever start solving cyber resilience (yes/no)?
"""

[[slides]]
source = """
# praha@pyvo.cz
- feedback 👂
- offer help 💪
- talk suggestions 👩‍🏫
"""
'''

if __name__ == "__main__":
    presentation = load_presentation(StringIO(CONTENT))
    slides = list(presentation.create_slides())
    app = PresentationApp(slides=slides, title=presentation.title or "")
    app.run()



import io
import json
import tomllib
from pathlib import Path
from typing import Iterable, Literal

from pydantic import BaseModel, Field

from clippt.slides import (
    CodeSlide,
    ErrorSlide,
    MarkdownSlide,
    PythonSlide,
    ShellSlide,
    Slide,
    load,
)


class SlideDescription(BaseModel):
    """Description of a slide."""

    class Config:
        extra = "forbid"

    type: Literal["python", "shell", "markdown", "code"] | None = None
    source: str | None = None
    path: Path | None = None
    """Path relative to the presentation."""

    title: str | None = None
    language: str | None = None
    """Language to be used for syntax highlighting."""

    alt_screen: bool | None = None
    mode: Literal["code", "output"] | None = None
    runnable: bool | None = None
    wait_for_key: bool | None = None

    classes: list[str] | None = None
    cwd: Path | None = None
    execute_before: str | None = None


class Presentation(BaseModel):
    """Description of a presentation."""

    slide_base_path: Path | None = Field(None, description="Base directory for slides")
    shell_cwd: Path | None = Field(
        None,
        description="Base directory for shell commands, relative to slide_base_path",
    )

    title: str | None = None
    slides: list[SlideDescription | str] = Field(default_factory=list)

    def _get_full_slide_path(self, path: Path) -> Path:
        if path.is_absolute():
            return path
        return (self.slide_base_path or Path(".")) / path

    @property
    def full_shell_cwd(self) -> Path | None:
        return (
            self._get_full_slide_path(self.shell_cwd)
            if self.shell_cwd
            else self.slide_base_path
        )

    def create_slides(self) -> Iterable[Slide]:
        for s in self.slides:
            if isinstance(s, str):
                slide_path = self._get_full_slide_path(Path(s))
                if slide_path.exists():
                    yield load(slide_path, cwd=self.full_shell_cwd)
                else:
                    yield ErrorSlide(source=f"Slide not found: {slide_path}")
            elif isinstance(s, SlideDescription):
                cwd = self._get_full_slide_path(s.cwd) if s.cwd else self.full_shell_cwd
                if s.path:
                    yield load(
                        path=self._get_full_slide_path(s.path),
                        **s.model_dump(
                            exclude_none=True, exclude={"type", "path", "cwd"}
                        ),
                    )
                else:
                    match s.type:
                        case "python":
                            yield PythonSlide(
                                **s.model_dump(exclude_none=True, exclude={"type"})
                            )
                        case "shell":
                            yield ShellSlide(
                                cwd=cwd,
                                **s.model_dump(
                                    exclude_none=True, exclude={"type", "cwd"}
                                ),
                            )
                        case "markdown" | None:
                            yield MarkdownSlide(
                                **s.model_dump(
                                    exclude_none=True,
                                    exclude={"type", "title", "language"},
                                )
                            )
                        case "code":
                            yield CodeSlide(
                                **s.model_dump(exclude_none=True, exclude={"type"})
                            )


def load_presentation(path_or_file: Path | str | io.TextIOBase, /) -> Presentation:
    if isinstance(path_or_file, io.TextIOBase):
        content = path_or_file.read()
        pwd = Path(".")
        data = tomllib.loads(content)
        return Presentation.model_validate(data | {"pwd": pwd})
    else:
        path = Path(path_or_file)
        pwd = path.parent

        match path.suffix.lower():
            case ".toml":
                data = tomllib.loads(path.read_text())
                return Presentation.model_validate(data | {"pwd": pwd})
            case ".json":
                data = json.loads(path.read_text())
                return Presentation.model_validate(data | {"pwd": pwd})
            case _:
                raise ValueError(f"Cannot parse {path}")



import io
import os
import subprocess
import traceback
from abc import ABC, abstractmethod
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from textwrap import dedent
from typing import Any, Callable, Final, Literal, Optional

import polars as pl
from pydantic import BaseModel, model_validator
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from shellingham import detect_shell
from textual.app import App
from textual.containers import VerticalScroll
from textual.widget import Widget
from textual.widgets import Markdown, Static
from textual_fastdatatable import DataTable
from textual_fastdatatable.backend import PolarsBackend

from clippt.utils import wait_for_key


class Slide(ABC, BaseModel):
    path: Path | None = None
    source: str = ""
    runnable: bool = False
    execute_before: str | None = None
    """Shell script to execute before the slide is rendered."""
    cwd: Path | None = None

    @model_validator(mode="after")
    def _load_on_start(self):
        self._load()
        return self

    def _load(self):
        if self.path:
            try:
                self.source = self.path.read_text(encoding="utf-8")
            except FileNotFoundError:
                self.source = f"File not found: {self.path}."

    def reload(self):
        self._load()

    def render(self, app: App) -> Widget:
        if self.execute_before:
            subprocess.run(
                self.execute_before.strip(),
                shell=True,
                capture_output=True,
                text=True,
                encoding="utf-8",
                cwd=self.cwd,
            )
        return self._render_impl(app)

    @abstractmethod
    def _render_impl(self, app: App) -> Widget: ...

    def run(self) -> None:
        pass


class CodeSlide(Slide):
    language: str | None = None
    title: Optional[str] = None
    is_title_markdown: bool = False

    def _render_impl(self, app) -> Widget:
        return self._render_code()

    def _render_code(self) -> VerticalScroll:
        if self.path:
            self.reload()
        code_lines = []
        for line in self.source.splitlines():
            line = line.rstrip()
            if "# HIDE_ABOVE" in line:
                code_lines = []
                continue
            if "# HIDE_BELOW" in line:
                break
            if "# HIDE" in line:
                continue
            code_lines.append(line)
        code = "\n".join(line for line in code_lines)
        if self.title:
            if self.is_title_markdown:
                md = Markdown(self.title + f"\n\n```{self.language}\n{code}\n```")
            else:
                md = Markdown(f"# {self.title}\n\n```{self.language}\n{code}\n```")
        else:
            md = Markdown(f"```{self.language}\n{code}\n```")
        return VerticalScroll(md, can_focus=False)


class ExecutableSlide(CodeSlide, ABC):
    """Slide with runnable code from external file or string."""

    mode: Literal["code", "output"] = "code"
    alt_screen: bool = False
    runnable: bool = True
    wait_for_key: bool = False
    is_error: bool = False

    _output: str | None = None

    def _load(self):
        self._output = None
        super()._load()

    def _render_impl(self, app) -> Widget:
        match self.mode:
            case "code":
                return self._render_code()
            case "output":
                if self.alt_screen:
                    self._exec_in_alternate_screen(app)
                    return self._render_code()
                else:
                    output = self._exec_inline(app)
                    return self._render_output(output=output, app=app)

    def _render_output(self, *, output: str, app: App) -> Widget:
        classes = "error" if self.is_error else "output"
        output_widget = Static(Text.from_ansi(output + "\n"), classes=classes)
        if self.title:
            if self.is_title_markdown:
                return VerticalScroll(
                    Markdown(self.title), output_widget, can_focus=False
                )
            return VerticalScroll(
                Markdown(f"# {self.title}"), output_widget, can_focus=False
            )
        return VerticalScroll(output_widget, can_focus=False)

    @abstractmethod
    def _exec_inline(self, app: App) -> str:
        """Execute the code and return the output."""

    @abstractmethod
    def _exec(self, app: App):
        """Execute the code."""

    def run(self):
        self.mode = "output" if self.mode == "code" else "code"

    def _exec_in_alternate_screen(self, app):
        with app.suspend():
            console = Console()
            console.clear()
            self._exec(app)
            if self.wait_for_key:
                wait_for_key()
            self.mode = "code"
            console.clear()


class PythonSlide(ExecutableSlide):
    language: Final[str] = "python"

    def _exec_inline(self, app) -> str:
        f = io.StringIO()
        with redirect_stdout(f):
            # TODO: Rethink? Remnant of old project, 99% not needed.
            # import plotext as plt
            # plt.plotsize(width=50, height=15)

            try:
                self._exec(app=app)
                return f.getvalue()
            except Exception as ex:
                out = StringIO()
                out.write(f"Error: {ex}\n")
                out.write("\n")
                traceback.print_exception(ex, file=out)
                return out.getvalue()

    def _exec(self, app: App) -> None:
        self.is_error = False
        try:
            exec(
                self.source,
                globals=globals()
                | {
                    "WIDTH": app.size.width - (10 if self.title else 4),
                    "HEIGHT": app.size.height - 2,
                },
            )
        except Exception:
            self.is_error = True
            raise

        # TODO: Rethink? Remnant of old project, 99% not needed.
        # import plotext as plt
        # plt.clear_figure()


class ShellSlide(ExecutableSlide):
    language: Final[str] = "shell"
    _executed: bool = False

    def __post_init__(self, **kwargs):
        if not self.source.strip():
            _, self.source = detect_shell()

    def _load(self):
        self._executed = False
        super()._load()

    def _exec(self, app: App, columns: int | None = None):
        env = None
        if columns is not None:
            env = {**os.environ, "COLUMNS": str(columns)}
        return subprocess.run(
            self.source.strip(),
            shell=True,
            capture_output=not self.alt_screen,
            text=True,
            encoding="utf-8",
            cwd=self.cwd,
            env=env,
        )

    def _exec_inline(self, app) -> str:
        if not self._executed:
            # margin: 0 3 adds 6 chars of horizontal chrome; subtract so the
            # command formats output to the exact displayable width.
            columns = app.size.width - 10
            with app.suspend():
                p = self._exec(app, columns=columns)
                self._output = p.stdout or p.stderr
                # TODO: Maybe we should raise if it fails and handle it elsewhere
                self.is_error = p.returncode != 0
                self._executed = True
        return self._output


class MarkdownSlide(Slide):
    classes: list[str] | None = None

    """Markdown slide with source from external file or string."""

    def _render_impl(self, app: App) -> Markdown:
        return Markdown(dedent(self.source), classes=" ".join(self.classes or []))


class TextSlide(Slide):
    title: Optional[str] = None

    def _render_impl(self, app: App) -> VerticalScroll:
        widgets = []
        if self.title:
            widgets.append(Markdown(f"# {self.title}"))
        widgets.append(Static(self.source))
        return VerticalScroll(*widgets, can_focus=False)


class FuncSlide(Slide):
    """Any slide created from a function."""

    f: Callable[[App], Markdown | Text | str]
    source: str = ""  # ignored
    path: None = None  # ignored

    def _render_impl(self, app: App) -> Widget:
        rendered = self.f(app)
        if isinstance(rendered, Widget):
            return rendered
        elif isinstance(rendered, str):
            return Markdown(dedent(rendered))
        elif isinstance(rendered, (Text, Panel)):
            return Static(rendered)
        else:
            raise NotImplementedError()


class DataSlide(Slide):
    data: Optional[pl.DataFrame] = None

    class Config:
        arbitrary_types_allowed = True

    def _render_impl(self, app: App) -> Widget:
        if self.data is not None:
            backend = PolarsBackend.from_dataframe(self.data)
            dt = DataTable(backend=backend, zebra_stripes=True, show_cursor=False)
            dt.can_focus = False
            return dt
        else:
            return Markdown("No data.")

    def _load(self) -> None:
        if self.path:
            match self.path.suffix:
                case ".csv":
                    self.data = pl.read_csv(self.path)
                case ".pq" | ".parquet":
                    self.data = pl.read_parquet(self.path)
                case _:
                    raise NotImplementedError()


class ErrorSlide(Slide):
    def _render_impl(self, app: App) -> Widget:
        return Static(Text.from_ansi(self.source), classes="error")


def slide(f: Callable[[App], Any]) -> FuncSlide:
    """Decorator to create a markdown slide from a function."""
    return FuncSlide(f=f)


def md(source: str, **kwargs) -> MarkdownSlide:
    """Helper function to create a simple Markdown slide."""
    return MarkdownSlide(path=None, source=source, **kwargs)


def py(source: str, **kwargs) -> ExecutableSlide:
    """Helper function to create a simple Python code slide."""
    return PythonSlide(path=None, source=source, **kwargs)


def sh(cmd, **kwargs) -> ExecutableSlide:
    """Helper function to create a shell command slide."""
    return ShellSlide(source=cmd, **kwargs)


def load(path: str | Path, **kwargs) -> Slide:
    """Load a slide from an external file."""
    path = Path(path)
    match path.suffix:
        case ".py":
            return PythonSlide(path=path, **kwargs)
        case ".md":
            return MarkdownSlide(path=path, **kwargs)
        case ".csv" | ".pq" | ".parquet":
            return DataSlide(path=path, **kwargs)
        case ".txt":
            return TextSlide(path=path, **kwargs)
        case other:
            language = kwargs.pop("language", EXT_LANGUAGE_MAPPING.get(other, "text"))
            return CodeSlide(path=path, language=language, **kwargs)


EXT_LANGUAGE_MAPPING = {
    ".json": "json",
    ".toml": "toml",
    ".yaml": "yaml",
    ".rs": "rust",
    ".scm": "scheme",
    ".go": "go",
}
